from .fragment_field import FRAGMENT_FIELD

FRAGMENT_FIELD_CONNECTION = '''
fragment fragmentFieldConnection on FieldConnection {
  totalCount
  edges {
    node {
     ...fragmentField
    }
  }
}
''' + FRAGMENT_FIELD



"""Commands (click)

"""
from flask.cli import with_appcontext
import click

import json

from .ops import define_tables
from .ops import export_db_to_dict_of_dict_list
from .ops import import_tables_from_csv_files
from .ops import export_db_to_csv_files
from .backup import backup_db


##__________________________________________________________________||
@click.command("init-db")
@with_appcontext
def init_db_command():
    """Clear existing data and create new tables."""
    define_tables()
    click.echo("Initialized the database.")


##__________________________________________________________________||
@click.command("dump-db")
@with_appcontext
def dump_db_command():
    """Dump the DB contents"""
    db_content = export_db_to_dict_of_dict_list()

    click.echo(json.dumps(db_content, indent=2, default=str))
    # https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable


##__________________________________________________________________||
@click.command("import-csv")
@click.argument("csvdir", type=click.Path(exists=True))
@with_appcontext
def import_csv_command(csvdir):
    """Import tables from CSV files in CSVDIR into the DB."""
    import_tables_from_csv_files(csvdir)


##__________________________________________________________________||
@click.command("export-csv")
@click.argument("csvdir", type=click.Path())
@with_appcontext
def export_csv_command(csvdir):
    """Export the tables into the DB to CSV files in CSVDIR."""
    export_db_to_csv_files(csvdir)


##__________________________________________________________________||
DEFAULT_EXCLUDES = ("github_tokens", "account_admins", "log")


@click.command("backup-db")
@click.option(
    "--exclude-csv",
    "-e",
    multiple=True,
    help=(
        f"A table to exclude from CSV backup. "
        f"Use multiple times to exclude multiple tables. "
        f'The tables {", ".join([f"{t!r}" for t in DEFAULT_EXCLUDES])} are excluded by default.'
    ),
)
@click.option(
    "--include-default-excludes",
    is_flag=True,
    help=(
        f"Include to CSV backup the tables that are excluded by default, "
        f'i.e., {", ".join([f"{t!r}" for t in DEFAULT_EXCLUDES])}.'
    ),
)
@with_appcontext
def backup_db_command(exclude_csv, include_default_excludes):
    """Back up the DB and its content as CSV to GitHub"""

    exclude_csv = set(exclude_csv)
    if not include_default_excludes:
        exclude_csv |= set(DEFAULT_EXCLUDES)
    backup_db(exclude_csv)


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[updateProductFilePath-immutableField] 1'] = {
    'data': {
        'product': {
            'paths': {
                'edges': [
                    {
                        'node': {
                            'note': '',
                            'path': 'nersc:/go/to/my/maps',
                            'product': {
                                'productId': '1001'
                            }
                        }
                    }
                ]
            }
        }
    }
}

snapshots['test_schema_success[updateProductFilePath] 1'] = {
    'data': {
        'updateProductFilePath': {
            'productFilePath': {
                'path': 'nersc:/go/to/my/new_product_v2'
            }
        }
    }
}

snapshots['test_schema_success[updateProductFilePath] 2'] = {
    'data': {
        'product': {
            'name': 'lat20190213',
            'note': '''- This is a dummy test with a lat map
- This should not depend on any beam''',
            'paths': {
                'edges': [
                    {
                        'node': {
                            'note': '- Note 1 updated',
                            'path': 'nersc:/go/to/my/new_product_v2',
                            'product': {
                                'productId': '1001'
                            }
                        }
                    }
                ]
            },
            'producedBy': 'pwg-pmn',
            'timePosted': '2019-02-13T10:05:23'
        }
    }
}



import pytest

from sqlalchemy import exc

from acondbs.models import Product, ProductRelation, ProductRelationType

from acondbs.db.sa import sa


##__________________________________________________________________||
def test_constraint(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # try to duplicate relation
    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").first()
        child1 = Product.query.filter_by(name="child1").first()
        type_ = ProductRelationType.query.filter_by(name="child").first()

        relation = ProductRelation()
        relation.type_ = type_
        relation.self_ = parent1
        relation.other = child1

        with pytest.raises(exc.IntegrityError):
            sa.session.commit()


##__________________________________________________________________||



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables


##__________________________________________________________________||
params = [
    [True, 200, "<!DOCTYPE html>"],
    [False, 400, "errors"],
]


@pytest.mark.parametrize("graphiql, code, data", params)
def test_disable(graphiql, code, data):
    app = create_app(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        ACONDBS_GRAPHIQL=graphiql,
        ACONDBS_GRAPHIQL_TEMPLATE_NO=2,
    )
    with app.app_context():
        define_tables()

    client = app.test_client()

    response = client.get("/graphql", headers={"Accept": "text/html"})
    assert code == response.status_code
    assert data in response.data.decode("utf-8")


##__________________________________________________________________||
params = [
    [None, "//cdn.jsdelivr.net/npm/graphiql@0.11.11/graphiql.min.js"],
    [1, "https://unpkg.com/graphiql/graphiql.min.js"],
    [2, "graphql-playground"],
]


@pytest.mark.parametrize("number, data", params)
def test_template(number, data):
    app = create_app(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        ACONDBS_GRAPHIQL=True,
        ACONDBS_GRAPHIQL_TEMPLATE_NO=number,
    )
    with app.app_context():
        define_tables()

    client = app.test_client()

    response = client.get("/graphql", headers={"Accept": "text/html"})
    assert 200 == response.status_code
    assert data in response.data.decode("utf-8")


##__________________________________________________________________||



import datetime

import pytest

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_users):

    y = app_users

    # Relation types:
    #   parent <-> child
    #   plaintiff <-> defendant

    # Products and Relations:
    #   map1 -> beam1
    #     |        |
    #     +--------+---> beam2
    #
    #   map2
    #   map3

    with y.app_context():
        ops.create_product_relation_type(
            type_={
                "type_id": 1,
                "name": "parent",
                "indef_article": "a",
                "singular": "parent",
                "plural": "parents",
            },
            reverse={
                "type_id": 2,
                "name": "child",
                "indef_article": "a",
                "singular": "child",
                "plural": "children",
            },
        )
        ops.create_product_relation_type(
            type_={
                "type_id": 3,
                "name": "plaintiff",
                "indef_article": "a",
                "singular": "plaintiff",
                "plural": "plaintiffs",
            },
            reverse={
                "type_id": 4,
                "name": "defendant",
                "indef_article": "a",
                "singular": "defendant",
                "plural": "defendants",
            },
        )
        ops.commit()

    with y.app_context():
        ops.create_field(
            field_id=1,
            name="contact",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=2,
            name="produced_by",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=3,
            name="date_produced",
            type_=ops.FieldType.Date,
        )
        ops.commit()

    with y.app_context():
        ops.create_product_type(
            type_id=1,
            name="map",
            order=2,
            indef_article="a",
            singular="map",
            plural="maps",
            icon="mdi-map",
            field_ids=[1, 2, 3],
        )
        ops.create_product_type(
            type_id=2,
            name="beam",
            order=1,
            indef_article="a",
            singular="beam",
            plural="beams",
            icon="mdi-spotlight-beam",
            field_ids=[1, 2, 3],
        )
        ops.commit()

    with y.app_context():
        ops.create_product(
            type_id=1,
            product_id=1,
            name="map1",
            date_produced=datetime.date(2020, 2, 1),
            attributes={3: datetime.date(2020, 2, 1)},
            paths=["site2:/another/way/map1"],
        )
        ops.create_product(
            type_id=1,
            product_id=2,
            name="map2",
            date_produced=datetime.date(2020, 2, 10),
            attributes={3: datetime.date(2020, 2, 10)},
            paths=["site1:/path/to/map2"],
        )
        ops.create_product(
            type_id=1,
            product_id=3,
            name="map3",
            date_produced=datetime.date(2020, 3, 19),
            attributes={3: datetime.date(2020, 3, 19)},
            paths=["site1:/path/to/map3", "site2:/another/way/map3"],
        )
        ops.create_product(
            type_id=2,
            product_id=4,
            name="beam1",
            date_produced=datetime.date(2020, 2, 5),
            attributes={3: datetime.date(2020, 2, 5)},
            paths=["site1:/path/to/beam1", "site2:/another/way/beam1"],
        )
        ops.create_product(
            type_id=2,
            product_id=5,
            name="beam2",
            date_produced=datetime.date(2020, 3, 4),
            attributes={3: datetime.date(2020, 3, 4)},
            paths=["site1:/path/to/beam2"],
        )
        ops.commit()

    with y.app_context():
        ops.create_product_file_path(
            product_id=1,
            path="site1:/path/to/map1",
            note="sample comment",
        )
        ops.commit()

    with y.app_context():
        ops.create_product_relation(
            type_id=1,
            self_product_id=4,
            other_product_id=1,
        )
        ops.create_product_relation(
            type_id=1,
            self_product_id=5,
            other_product_id=1,
        )
        ops.create_product_relation(
            type_id=1,
            self_product_id=5,
            other_product_id=4,
        )
        ops.commit()

    yield y

##__________________________________________________________________||



from acondbs.models import Product


##__________________________________________________________________||
def test_relations(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").first()
        child1 = Product.query.filter_by(name="child1").first()
        child2 = Product.query.filter_by(name="child2").first()

        assert 2 == len(parent1.relations)
        assert 1 == len(child1.relations)
        assert 1 == len(child2.relations)

        assert "child" == parent1.relations[0].type_.name
        assert "child" == parent1.relations[1].type_.name
        assert "parent" == child1.relations[0].type_.name
        assert "parent" == child2.relations[0].type_.name

        assert child1 is parent1.relations[0].other
        assert child2 is parent1.relations[1].other

        assert parent1 is child1.relations[0].other
        assert parent1 is child2.relations[0].other

        assert parent1.relations[0] is child1.relations[0].reverse
        assert parent1.relations[0].reverse is child1.relations[0]

        assert parent1.relations[1] is child2.relations[0].reverse
        assert parent1.relations[1].reverse is child2.relations[0]


##__________________________________________________________________||



"""declare ORM models for GitHub related tables
"""

##__________________________________________________________________||
from .github_token import GitHubToken  # noqa: F401
from .github_org import GitHubOrg  # noqa: F401
from .github_user import GitHubUser  # noqa: F401
from .github_org_membership import GitHubOrgMembership  # noqa: F401

##__________________________________________________________________||



import datetime
from sqlalchemy.orm import aliased

import pytest

from acondbs.db.sa import sa

from acondbs.models import (
    Product,
    ProductType,
    FieldType,
    Field,
    TypeFieldAssociation,
    AttributeUnicodeText,
    AttributeBoolean,
    AttributeInteger,
    AttributeFloat,
    AttributeDate,
    AttributeDateTime,
    AttributeTime
)


##__________________________________________________________________||
params = [
    pytest.param(FieldType.UnicodeText, AttributeUnicodeText, "value1 値１", id='UnicodeText'),  # fmt: skip
    pytest.param(FieldType.Boolean, AttributeBoolean, True, id='Boolean'),  # fmt: skip
    pytest.param(FieldType.Integer, AttributeInteger, -512442, id='Integer'),  # fmt: skip
    pytest.param(FieldType.Float, AttributeFloat, 3.14592613, id='Float'),  # fmt: skip
    pytest.param(FieldType.Date, AttributeDate, datetime.date(2020, 2, 1), id='Date'),  # fmt: skip
    pytest.param(FieldType.DateTime, AttributeDateTime, datetime.datetime(2020, 2, 1, 9, 10, 25), id='DateTime'),  # fmt: skip
    pytest.param(FieldType.Time, AttributeTime, datetime.time(9, 10, 25), id='Time'),  # fmt: skip
]


@pytest.mark.parametrize('field_type, AttributeClass, value', params)
def test_repr(app_empty, field_type, AttributeClass, value):
    app = app_empty  # noqa: F841

    attr1 = AttributeClass(value=value)
    repr(attr1)

    attr1.field = Field(name="attr1", type_=field_type)
    repr(attr1)


@pytest.mark.parametrize('field_type, AttributeClass, value', params)
def test_obj(app_empty, field_type, AttributeClass, value):
    app = app_empty  # noqa: F841
    field = Field(name="field1", type_=field_type)
    attr = AttributeClass(field=field, value=value)  # noqa: F841


@pytest.mark.parametrize('field_type, AttributeClass, value', params)
def test_commit(app_empty, field_type, AttributeClass, value):
    app = app_empty
    with app.app_context():
        field = Field(name="field1", type_=field_type)
        product_type = ProductType(name="map")
        product_type.fields = [TypeFieldAssociation(field=field)]
        sa.session.add(product_type)
        sa.session.commit()

    with app.app_context():
        product_type = ProductType.query.filter_by(name="map").one()
        field = product_type.fields[0].field
        product = Product(name="product1", type_=product_type)
        attr = AttributeClass(product=product, field=field, value=value)
        sa.session.add(attr)
        sa.session.commit()

    with app.app_context():
        attr = AttributeClass.query.one()
        assert attr.__class__ is AttributeClass
        assert attr.field.name == 'field1'
        assert attr.field.type_ is field_type
        assert attr.value == value
        assert getattr(attr.product, AttributeClass.backref_column) == [attr]
        assert getattr(attr.field, AttributeClass.backref_column) == [attr]


##__________________________________________________________________||
def test_filter(app):

    with app.app_context():
        expected = Product.query.filter_by(product_id=1).one()
        query = (
            Product.query.join(AttributeUnicodeText).join(Field)
            .filter(Field.name == "attr1", AttributeUnicodeText.value == "value1")
        )
        actual = query.one()
        assert actual is expected

        # another way
        query = (
            Product.query
            .join(AttributeUnicodeText)
            .filter_by(value="value1")
            .join(Field)
            .filter_by(name="attr1")
        )
        actual = query.one()
        assert actual is expected


def test_order(app):

    with app.app_context():
        map1 = Product.query.filter_by(product_id=1).one()
        map2 = Product.query.filter_by(product_id=2).one()
        map3 = Product.query.filter_by(product_id=3).one()

        expected = [map3, map2, map1]

        query = (
            Product.query.join(ProductType)
            .filter_by(name="map")
            .join(AttributeDate).join(Field)
            .filter_by(name="date_produced")
            .order_by(AttributeDate.value.desc())
        )

        print(query)

        actual = query.all()

        assert actual == expected


##__________________________________________________________________||
def test_order_nested(app_empty):
    app = app_empty

    field_attr1 = Field(name="attr1", type_=FieldType.UnicodeText)
    field_attr2 = Field(name="attr2", type_=FieldType.UnicodeText)
    Map = ProductType(type_id=1, name="map")
    Map.fields = [
        TypeFieldAssociation(field=field_attr1),
        TypeFieldAssociation(field=field_attr2),
    ]

    map1 = Product(product_id=1, name="map1", type_=Map)
    map2 = Product(product_id=2, name="map2", type_=Map)
    map3 = Product(product_id=3, name="map3", type_=Map)
    map4 = Product(product_id=4, name="map4", type_=Map)
    map5 = Product(product_id=5, name="map5", type_=Map)

    AttributeUnicodeText(product=map1, field=field_attr1, value="1")
    AttributeUnicodeText(product=map2, field=field_attr1, value="2")
    AttributeUnicodeText(product=map3, field=field_attr1, value="1")
    AttributeUnicodeText(product=map4, field=field_attr1, value="2")
    AttributeUnicodeText(product=map5, field=field_attr1, value="1")

    AttributeUnicodeText(product=map1, field=field_attr2, value="b")
    AttributeUnicodeText(product=map2, field=field_attr2, value="a")
    AttributeUnicodeText(product=map3, field=field_attr2, value="c")
    AttributeUnicodeText(product=map4, field=field_attr2, value="b")
    AttributeUnicodeText(product=map5, field=field_attr2, value="a")

    with app.app_context():
        sa.session.add(Map)
        sa.session.commit()

    with app.app_context():

        map1 = Product.query.filter_by(product_id=1).one()
        map2 = Product.query.filter_by(product_id=2).one()
        map3 = Product.query.filter_by(product_id=3).one()
        map4 = Product.query.filter_by(product_id=4).one()
        map5 = Product.query.filter_by(product_id=5).one()

        expected = [map2, map4, map5, map1, map3]

        # ORM Alias: https://docs.sqlalchemy.org/en/14/tutorial/data_select.html#orm-entity-aliases
        # refer to the same table multiple times
        AliasedAttributeUnicodeText1 = aliased(AttributeUnicodeText)
        AliasedAttributeUnicodeText2 = aliased(AttributeUnicodeText)

        AliasedField1 = aliased(Field)
        AliasedField2 = aliased(Field)

        # sort descending order of attr1 then ascending order of attr2
        query = (
            Product.query.join(ProductType)
            .filter_by(name="map")
            .join(AliasedAttributeUnicodeText1)
            .join(AliasedField1)
            .filter_by(name="attr1")
            .order_by(AliasedAttributeUnicodeText1.value.desc())
            .join(AliasedAttributeUnicodeText2, Product.attributes_unicode_text)
            .join(AliasedField2)
            .filter_by(name="attr2")
            .order_by(AliasedAttributeUnicodeText2.value)
        )

        print(query)

        actual = query.all()

        assert actual == expected


##__________________________________________________________________||
def test_delte_orphan(app):

    with app.app_context():
        map1 = Product.query.filter_by(product_id=1).one()
        sa.session.delete(map1)
        sa.session.commit()

    with app.app_context():
        map1 = Product.query.filter_by(product_id=1).one_or_none()
        assert map1 is None
        attr = AttributeUnicodeText.query.filter_by(
            name="attr1", value="value1"
        ).one_or_none()
        assert attr is None


##__________________________________________________________________||



import datetime

from acondbs.db.sa import sa
from acondbs.models import GitHubToken, GitHubUser


##__________________________________________________________________||
def test_query(app):

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        token1, token2 = user1.tokens

        assert token1.token == "token_001"
        assert token1.scope == "read:org"
        assert token1.user == user1

        assert token2.token == "token_002"
        assert token2.scope == "read:org"
        assert token2.user == user1


##__________________________________________________________________||
def test_add(app):

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        ntokens = len(user1.tokens)  # = 2
        token3 = GitHubToken(token="token_003", scope="read:org", user=user1)
        sa.session.commit()

    with app.app_context():
        token3 = (
            GitHubToken.query.join(GitHubUser)
            .filter(GitHubUser.login == "octocat")
            .filter(GitHubToken.token == "token_003")
            .one()
        )
        assert "token_003" == token3.token
        assert "read:org" == token3.scope
        assert (
            datetime.datetime(2021, 1, 4, 14, 32, 20) == token3.time_created
        )  # default value set in the fixture mock_datetime
        user1 = token3.user
        assert "octocat" == user1.login
        assert ntokens + 1 == len(user1.tokens)


##__________________________________________________________________||
def test_delete(app):

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        ntokens = len(user1.tokens)  # = 2
        token1 = user1.tokens[0]
        sa.session.delete(token1)
        sa.session.commit()

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        assert ntokens - 1 == len(user1.tokens)  # 1

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        token2 = user1.tokens[0]
        sa.session.delete(token2)
        sa.session.commit()

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        assert ntokens - 2 == len(user1.tokens)  # 0


##__________________________________________________________________||
def test_delete_cascade(app):

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        sa.session.delete(user1)
        sa.session.commit()

    with app.app_context():
        tokens = GitHubToken.query.all()
        assert 0 == len(tokens)


##__________________________________________________________________||



import pytest

from ..funcs import assert_query

QUERY = """
{ webConfig {
    configId
    headTitle
    toolbarTitle
    devtoolLoadingstate
    productCreationDialog
    productUpdateDialog
    productDeletionDialog
  }
}
"""


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": QUERY},
        id="query",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data)

##__________________________________________________________________||



import pytest

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_users):
    y = app_users
    with y.app_context():
        ops.create_field(
            field_id=1,
            name="contact",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=2,
            name="produced_by",
            type_=ops.FieldType.UnicodeText,
        )
        ops.create_field(
            field_id=3,
            name="date_produced",
            type_=ops.FieldType.Date,
        )
        ops.commit()
    yield y


##__________________________________________________________________||



"""Assert owners specified in the config file are added as admins in DB
"""

from pathlib import Path
import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables
from acondbs.db.sa import sa
from acondbs.models import AccountAdmin

from ..constants import SAMPLE_DIR


##__________________________________________________________________||
def test_no_error_in_create_app():
    """Assert an error does not occur in create_app() when tables are not defined"""
    config_path = Path(SAMPLE_DIR, "config.py")
    kwargs = dict(
        SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
        ACONDBS_OWNERS_GITHUB_LOGINS="octocat,dojocat",
    )
    app = create_app(config_path, **kwargs)  # noqa: F841


##__________________________________________________________________||
params = [
    [{"ACONDBS_OWNERS_GITHUB_LOGINS": ""}, {"octocat"}],
    [{"ACONDBS_OWNERS_GITHUB_LOGINS": ","}, {"octocat"}],
    [{"ACONDBS_OWNERS_GITHUB_LOGINS": "octocat"}, {"octocat"}],
    [{"ACONDBS_OWNERS_GITHUB_LOGINS": "octocat  "}, {"octocat"}],
    [{"ACONDBS_OWNERS_GITHUB_LOGINS": "dojocat"}, {"octocat", "dojocat"}],
    [
        {"ACONDBS_OWNERS_GITHUB_LOGINS": "octocat,dojocat"},
        {"octocat", "dojocat"},
    ],
    [
        {"ACONDBS_OWNERS_GITHUB_LOGINS": "octocat  ,  dojocat"},
        {"octocat", "dojocat"},
    ],
    [
        {"ACONDBS_OWNERS_GITHUB_LOGINS": "octocat  ,  dojocat,, ,  octocat  "},
        {"octocat", "dojocat"},
    ],
]


@pytest.mark.parametrize("kwargs, expected", params)
def test_add_owners_to_db_as_admins(kwargs, expected, tmpdir_factory):
    config_path = Path(SAMPLE_DIR, "config.py")

    # use a file because, with memory, a DB will be recreated in the
    # second create_app().
    tmpdir = str(tmpdir_factory.mktemp("models"))
    tmp_database_path = Path(tmpdir, "product.sqlite3")
    kwargs.update(
        dict(SQLALCHEMY_DATABASE_URI=f"sqlite:///{tmp_database_path}")
    )

    # define tables and add a admin
    app_ = create_app(config_path, **kwargs)
    with app_.app_context():
        define_tables()

        octocat = AccountAdmin(git_hub_login="octocat")
        sa.session.add(octocat)
        sa.session.commit()

    app = create_app(config_path, **kwargs)
    with app.app_context():
        assert expected == {e.git_hub_login for e in AccountAdmin.query.all()}


##__________________________________________________________________||



import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, FieldType, Field, TypeFieldAssociation


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):

    y = app_empty

    field1 = Field(name="field1", type_=FieldType.UnicodeText)
    field2 = Field(name="field2", type_=FieldType.Integer)

    type1 = ProductType(  # noqa: F841
        name="type1",
        fields=[
            TypeFieldAssociation(field=field1),
            TypeFieldAssociation(field=field2),
        ],
    )

    with y.app_context():
        sa.session.add(field1)
        sa.session.add(field2)
        sa.session.add(type1)
        sa.session.commit()

    yield y


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_table_names 1'] = [
    'product_types',
    'products',
    'product_file_paths',
    'product_relation_types',
    'product_relations',
    'attribute_unicode_text',
    'attribute_boolean',
    'attribute_integer',
    'attribute_float',
    'attribute_date',
    'attribute_date_time',
    'attribute_time',
    'field',
    'type_field_association',
    'github_tokens',
    'github_orgs',
    'github_users',
    'github_org_memberships',
    'account_admins',
    'web_config',
    'log'
]



import datetime

from ...db.sa import sa


##__________________________________________________________________||
class Product(sa.Model):
    # TODO: rename the class name to be more generic, e.g., "Entry"
    # TODO: use singular for the table name
    __tablename__ = "products"
    product_id = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(sa.ForeignKey("product_types.type_id"), nullable=False)
    type_ = sa.relationship("ProductType", backref=sa.backref("products"))
    name = sa.Column(sa.Text(), nullable=False)
    contact = sa.Column(sa.Text())  # TODO: move to attribute model
    date_produced = sa.Column(sa.Date())  # TODO: move to attribute model
    produced_by = sa.Column(sa.Text())  # TODO: move to attribute model
    time_posted = sa.Column(
        sa.DateTime(), default=lambda: datetime.datetime.now()
    )
    posted_by = sa.Column(sa.Text())
    posting_git_hub_user_id = sa.Column(sa.ForeignKey("github_users.user_id"))
    posting_git_hub_user = sa.relationship(
        "GitHubUser",
        foreign_keys=[posting_git_hub_user_id],
        backref=sa.backref("posted_products", cascade="all"),
    )
    time_updated = sa.Column(sa.DateTime())
    updated_by = sa.Column(sa.Text())
    updating_git_hub_user_id = sa.Column(sa.ForeignKey("github_users.user_id"))
    updating_git_hub_user = sa.relationship(
        "GitHubUser",
        foreign_keys=[updating_git_hub_user_id],
        backref=sa.backref("updated_products", cascade="all"),
    )
    note = sa.Column(sa.Text())
    __table_args__ = (
        sa.UniqueConstraint("type_id", "name", name="_type_id_name"),
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import ProductType, Product

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.


##__________________________________________________________________||
def test_simple(app):
    """A simple test of adding an object"""

    with app.app_context():

        # the number of the product types is zero initially
        assert 0 == len(ProductType.query.all())

    # this instantiation doesn't need be within a app context
    type1 = ProductType(name="type1")

    with app.app_context():
        sa.session.add(type1)
        sa.session.commit()

    with app.app_context():

        # the number of the product types is increased by one
        assert 1 == len(ProductType.query.all())

        # the new product type can be retrieved in a different app context
        type1_ = ProductType.query.filter_by(name="type1").one_or_none()
        assert isinstance(type1_, ProductType)


##__________________________________________________________________||
def test_python_object(app):
    """A simple test about Python object"""

    type1 = ProductType(name="type1")

    with app.app_context():
        sa.session.add(type1)
        sa.session.commit()

        type1_ = ProductType.query.filter_by(name="type1").first()

        # the query returns the same Python object
        assert type1 is type1_

    with app.app_context():
        type1_ = ProductType.query.filter_by(name="type1").first()

        # In a different app context, no longer the same Python object
        assert type1 is not type1_


##__________________________________________________________________||
def test_primary_key(app):
    """A simple test about the primary key"""

    type1 = ProductType(name="type1")

    # The primary key (type_id) is None at this point
    assert type1.type_id is None

    with app.app_context():
        sa.session.add(type1)
        sa.session.commit()

        # After the commit, type_id is automatically assigned
        type_id = type1.type_id
        assert type_id is not None

    with app.app_context():

        # The object can be retrived by the product_id in another context
        type1 = ProductType.query.filter_by(type_id=type_id).first()
        assert "type1" == type1.name


##__________________________________________________________________||
def test_relation(app):
    """A simple test of adding an object with relation"""

    type1 = ProductType(name="type1")
    product1 = Product(name="product1", type_=type1)

    # The relation has been already established
    assert type1 is product1.type_
    assert [product1] == type1.products

    # The primary and foreign keys are still None
    assert type1.type_id is None
    assert product1.product_id is None
    assert product1.type_id is None

    with app.app_context():
        sa.session.add(product1)
        sa.session.commit()

        # The primary keys are assigned
        assert type1.type_id is not None
        assert product1.product_id is not None

        # The foreign key is correctly set
        assert product1.type_id == type1.type_id

    with app.app_context():
        product1 = Product.query.filter_by(name="product1").first()

        # The relation is preserved in a different app context
        type1 = product1.type_
        assert "type1" == type1.name
        assert product1 is type1.products[0]
        assert product1.type_id == type1.type_id


##__________________________________________________________________||



"""re-initialize migration

Revision ID: cc64c8ef2115
Revises: 
Create Date: 2020-05-20 19:58:21.088239

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'cc64c8ef2115'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('product_relation_types',
    sa.Column('type_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('type_id')
    )
    op.create_index(op.f('ix_product_relation_types_name'), 'product_relation_types', ['name'], unique=True)
    op.create_table('product_types',
    sa.Column('type_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('type_id')
    )
    op.create_index(op.f('ix_product_types_name'), 'product_types', ['name'], unique=True)
    op.create_table('products',
    sa.Column('product_id', sa.Integer(), nullable=False),
    sa.Column('type_id', sa.Integer(), nullable=True),
    sa.Column('name', sa.Text(), nullable=False),
    sa.Column('contact', sa.Text(), nullable=True),
    sa.Column('date_produced', sa.Date(), nullable=True),
    sa.Column('produced_by', sa.Text(), nullable=True),
    sa.Column('date_posted', sa.Date(), nullable=True),
    sa.Column('posted_by', sa.Text(), nullable=True),
    sa.Column('date_updated', sa.Date(), nullable=True),
    sa.Column('updated_by', sa.Text(), nullable=True),
    sa.Column('note', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['type_id'], ['product_types.type_id'], ),
    sa.PrimaryKeyConstraint('product_id'),
    sa.UniqueConstraint('type_id', 'name', name='_type_id_name')
    )
    op.create_table('product_file_paths',
    sa.Column('path_id', sa.Integer(), nullable=False),
    sa.Column('path', sa.Text(), nullable=True),
    sa.Column('note', sa.Text(), nullable=True),
    sa.Column('product_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['product_id'], ['products.product_id'], ),
    sa.PrimaryKeyConstraint('path_id')
    )
    op.create_table('product_relations',
    sa.Column('relation_id', sa.Integer(), nullable=False),
    sa.Column('type_id', sa.Integer(), nullable=True),
    sa.Column('self_product_id', sa.Integer(), nullable=True),
    sa.Column('other_product_id', sa.Integer(), nullable=True),
    sa.Column('reverse_relation_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['other_product_id'], ['products.product_id'], ),
    sa.ForeignKeyConstraint(['reverse_relation_id'], ['product_relations.relation_id'], ),
    sa.ForeignKeyConstraint(['self_product_id'], ['products.product_id'], ),
    sa.ForeignKeyConstraint(['type_id'], ['product_relation_types.type_id'], ),
    sa.PrimaryKeyConstraint('relation_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('product_relations')
    op.drop_table('product_file_paths')
    op.drop_table('products')
    op.drop_index(op.f('ix_product_types_name'), table_name='product_types')
    op.drop_table('product_types')
    op.drop_index(op.f('ix_product_relation_types_name'), table_name='product_relation_types')
    op.drop_table('product_relation_types')
    # ### end Alembic commands ###



"""constraint on product_relations

Revision ID: ce157c2fa22e
Revises: 0f7fd06aa729
Create Date: 2020-05-26 18:48:17.865643

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ce157c2fa22e'
down_revision = '0f7fd06aa729'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relations', schema=None) as batch_op:
        batch_op.create_unique_constraint('_type_id_self_product_id_other_product_id', ['type_id', 'self_product_id', 'other_product_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relations', schema=None) as batch_op:
        batch_op.drop_constraint('_type_id_self_product_id_other_product_id', type_='unique')

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[one] 1'] = {
    'data': {
        'allFields': {
            'edges': [
                {
                    'node': {
                        'fieldId': '1',
                        'name': 'contact',
                        'type_': 'UNICODE_TEXT'
                    }
                },
                {
                    'node': {
                        'fieldId': '2',
                        'name': 'produced_by',
                        'type_': 'UNICODE_TEXT'
                    }
                },
                {
                    'node': {
                        'fieldId': '3',
                        'name': 'date_produced',
                        'type_': 'DATE'
                    }
                }
            ],
            'totalCount': 3
        }
    }
}



import graphene
from graphql import GraphQLError

from ...db.sa import sa
from ...db.backup import request_backup_db

from ...models import (
    GitHubToken as GitHubTokenModel,
)
from ...github.ops import (
    update_org_member_lists,
    add_org,
    delete_org,
    store_token_for_code,
    authenticate,
)

from . import type_


##__________________________________________________________________||
class AddGitHubOrg(graphene.Mutation):
    class Arguments:
        login = graphene.String(required=True)

    ok = graphene.Boolean()
    git_hub_org = graphene.Field(lambda: type_.GitHubOrg)

    def mutate(root, info, login):
        model = add_org(login)
        ok = True
        # request_backup_db()
        return AddGitHubOrg(git_hub_org=model, ok=ok)


class DeleteGitHubOrg(graphene.Mutation):
    class Arguments:
        login = graphene.String(required=True)

    ok = graphene.Boolean()

    def mutate(root, info, login):
        delete_org(login)
        ok = True
        # request_backup_db()
        return DeleteGitHubOrg(ok=ok)


##__________________________________________________________________||
class AuthenticateWithGitHub(graphene.Mutation):
    class Arguments:
        code = graphene.String(required=True)

    authPayload = graphene.Field(lambda: type_.AuthPayload)

    def mutate(root, info, code):
        token_dict = authenticate(code)
        if not token_dict:
            raise GraphQLError("Unsuccessful to obtain the token")
        authPayload = type_.AuthPayload(token=token_dict["access_token"])
        return AuthenticateWithGitHub(authPayload=authPayload)


##__________________________________________________________________||
class AddGitHubAdminAppToken(graphene.Mutation):
    """Add a token for a GitHub Admin App"""

    class Arguments:
        code = graphene.String(required=True)

    ok = graphene.Boolean()

    def mutate(root, info, code):
        store_token_for_code(code)
        ok = True
        request_backup_db()
        return AddGitHubAdminAppToken(ok=ok)


##__________________________________________________________________||
class DeleteGitHubAdminAppToken(graphene.Mutation):
    """Delete a token for a GitHub Admin App"""

    class Arguments:
        token_id = graphene.Int(required=True)

    ok = graphene.Boolean()

    def mutate(root, info, token_id):
        model = GitHubTokenModel.query.filter_by(token_id=token_id).one()
        sa.session.delete(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteGitHubAdminAppToken(ok=ok)


##__________________________________________________________________||
class UpdateGitHubOrgMemberLists(graphene.Mutation):
    """Update the member lists of GitHub organizations"""

    ok = graphene.Boolean()

    def mutate(root, info):
        update_org_member_lists()
        ok = True
        # request_backup_db()
        return UpdateGitHubOrgMemberLists(ok=ok)


##__________________________________________________________________||



FRAGMENT_PRODUCT = """
fragment fragmentProduct on Product {
  productId
  typeId
  type_ {
    typeId
    name
  }
  attributesUnicodeText {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  attributesBoolean {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  attributesInteger {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  attributesFloat {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  attributesDate {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  attributesDateTime {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  attributesTime {
    edges {
      node {
        name
        field {
          name
          type_
        }
        value
      }
    }
  }
  name
  contact
  dateProduced
  producedBy
  timePosted
  postedBy
  postingGitHubUser {
    login
  }
  timeUpdated
  updatedBy
  updatingGitHubUser {
    login
  }
  paths {
    edges {
      node {
        pathId
        path
        note
      }
    }
  }
  relations {
    edges {
      node {
        relationId
        typeId
        type_ {
          typeId
          name
        }
        otherProductId
        other {
          productId
          typeId
          type_ {
            typeId
            name
          }
          name
        }
        reverseRelationId
        reverse {
          relationId
          typeId
          type_ {
            typeId
            name
          }
        }
      }
    }
  }
  note
}
"""



"""add table log

Revision ID: 3f804a98ab06
Revises: 35e6ddccd22a
Create Date: 2021-10-22 19:06:30.269565

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3f804a98ab06'
down_revision = '35e6ddccd22a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('log',
    sa.Column('id_', sa.Integer(), nullable=False),
    sa.Column('level', sa.Text(), nullable=True),
    sa.Column('message', sa.Text(), nullable=True),
    sa.Column('time', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id_', name=op.f('pk_log'))
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('log')
    # ### end Alembic commands ###



from sqlalchemy import exc

import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, FieldType, Field, TypeFieldAssociation


##__________________________________________________________________||
def test_repr(app_empty):
    app = app_empty  # noqa: F841

    assoc = TypeFieldAssociation()
    repr(assoc)

    field1 = Field(name="field1", type_=FieldType.UnicodeText)
    type1 = ProductType(name="type1")

    assoc.type_ = type1
    assoc.field = field1
    repr(assoc)


##__________________________________________________________________||
def test_relationship(app):
    with app.app_context():
        field1 = Field.query.filter_by(name="field1").one()
        field2 = Field.query.filter_by(name="field2").one()
        type1 = ProductType.query.filter_by(name="type1").one()
        association1 = TypeFieldAssociation.query.filter_by(type_=type1, field=field1).one()  # fmt: skip
        association2 = TypeFieldAssociation.query.filter_by(type_=type1, field=field2).one()  # fmt: skip
        assert type1.fields == [association1, association2]
        assert field1.entry_types == [association1]
        assert field2.entry_types == [association2]
        assert association1.type_ is type1
        assert association1.field is field1
        assert association2.type_ is type1
        assert association2.field is field2


##__________________________________________________________________||
def test_cascade_deleting_type(app):
    """test delete a type

    When a type is deleted, its associations should be automatically
    deleted while the fields themselves should still exist.

    """

    with app.app_context():
        type1 = ProductType.query.filter_by(name="type1").one()
        sa.session.delete(type1)
        sa.session.commit()

    with app.app_context():
        assert ProductType.query.filter_by(name="type1").one_or_none() is None
        assert Field.query.filter_by(name="field1").one()
        assert Field.query.filter_by(name="field2").one()
        associations = TypeFieldAssociation.query.all()
        assert not associations  # emtpy


def test_cascade_updating_type(app):
    with app.app_context():
        type1 = ProductType.query.filter_by(name="type1").one()

        type1.fields = [f for f in type1.fields if f.field.name == "field2"]
        # remove field1

        sa.session.commit()

    with app.app_context():
        field1 = Field.query.filter_by(name="field1").one()
        field2 = Field.query.filter_by(name="field2").one()
        type1 = ProductType.query.filter_by(name="type1").one()
        association2 = TypeFieldAssociation.query.filter_by(type_=type1, field=field2).one()  # fmt: skip
        assert type1.fields == [association2]
        assert field1.entry_types == []
        assert field2.entry_types == [association2]
        assert association2.type_ is type1
        assert association2.field is field2
        assert type1.fields == [association2]
        assert TypeFieldAssociation.query.all() == [association2]
        # association1 is deleted from the DB


##__________________________________________________________________||
def test_nullable_deleting_field(app):
    """test delete a field

    A field cannot be deleted if it is associated

    """

    with app.app_context():
        field1 = Field.query.filter_by(name="field1").one()
        sa.session.delete(field1)
        with pytest.raises(AssertionError):
            sa.session.commit()

    with app.app_context():
        field1 = Field.query.filter_by(name="field1").one()
        assert field1.entry_types


##__________________________________________________________________||
def test_unique_constraint(app):
    # type_id and field_id are the primary keys.
    # A type cannot have multiple same field.
    with app.app_context():
        field1 = Field.query.filter_by(name="field1").one()
        type2 = ProductType(
            name="type2",
            fields=[
                TypeFieldAssociation(field=field1),
                TypeFieldAssociation(field=field1),
            ],
        )
        sa.session.add(type2)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()


##__________________________________________________________________||



from acondbs.models import Product, ProductRelation, ProductRelationType


##__________________________________________________________________||
def test_example_how_to_query(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    with app.app_context():

        parent1 = Product.query.filter_by(name="parent1").first()
        child1 = Product.query.filter_by(name="child1").first()
        child2 = Product.query.filter_by(name="child2").first()

        # Product relations with the type name "child"
        relations = (
            ProductRelation.query.join(ProductRelationType)
            .filter(ProductRelationType.name == "child")
            .all()
        )
        assert ["child", "child"] == [r.type_.name for r in relations]

        # Products with "parent"
        products = (
            Product.query.join(
                ProductRelation,
                (Product.product_id == ProductRelation.self_product_id),
            )
            .join(ProductRelationType)
            .filter(ProductRelationType.name == "parent")
            .all()
        )
        assert [child1, child2] == products

        # Products with "child"
        products = (
            Product.query.join(
                ProductRelation,
                (Product.product_id == ProductRelation.self_product_id),
            )
            .join(ProductRelationType)
            .filter(ProductRelationType.name == "child")
            .all()
        )
        assert [parent1] == products


##__________________________________________________________________||



from .fragment_product import FRAGMENT_PRODUCT

FRAGMENT_PRODUCT_CONNECTION = '''
fragment fragmentProductConnection on ProductConnection {
  edges {
    node {
     ...fragmentProduct
    }
  }
}
''' + FRAGMENT_PRODUCT



import graphene

from .. import type_
from .... import ops


##__________________________________________________________________||
class CreateProductRelationInput(graphene.InputObjectType):
    """An input to createProductRelation()"""

    type_id = graphene.Int(
        required=True,
        description=(
            "The typeId of the product relation type of the relation "
            'from "self" to the "other"'
        ),
    )
    self_product_id = graphene.Int(
        required=True, description=("The productId of the self product")
    )
    other_product_id = graphene.Int(
        required=True, description=("The productId of the other product")
    )


##__________________________________________________________________||
class CreateProductRelation(graphene.Mutation):
    """Add relations between two products. The arguments only specify the relation
    from one product to the other. The reverse relation will be also added.

    """

    class Arguments:
        input = CreateProductRelationInput(required=True)

    ok = graphene.Boolean()
    product_relation = graphene.Field(lambda: type_.ProductRelation)

    def mutate(root, info, input):
        model = ops.create_product_relation(**input)
        ops.commit()
        ok = True
        return CreateProductRelation(product_relation=model, ok=ok)


class DeleteProductRelation(graphene.Mutation):
    """Remove relations from two products."""

    class Arguments:
        relation_id = graphene.Int(
            required=True,
            description=(
                "The relationId of a relation. The reverse relation "
                "will automatically be removed."
            ),
        )

    ok = graphene.Boolean()

    def mutate(root, info, relation_id):
        ops.delete_product_relation(relation_id)
        ops.commit()
        ok = True
        return DeleteProductRelation(ok=ok)


##__________________________________________________________________||



from flask import json

QUERY = """{
  webConfig {
    headTitle
    toolbarTitle
  }
}"""


##__________________________________________________________________||
def test_graphql(client, snapshot):
    path = "/graphql"

    response = client.get(path, query_string=dict(query=QUERY))
    assert 200 == response.status_code

    un_jsonified = json.loads(response.data)
    snapshot.assert_match(un_jsonified)
    # print(json.dumps(un_jsonified, indent=2))


##__________________________________________________________________||



from acondbs import ops

from acondbs.models import ProductRelation


def test_create(app):
    with app.app_context():
        count = ProductRelation.query.count()
        model = ops.create_product_relation(
            type_id=1, self_product_id=1, other_product_id=4
        )
        ops.commit()
        relation_id = model.relation_id
        assert relation_id

    with app.app_context():
        assert ProductRelation.query.count() == (count + 2)
        model = ProductRelation.query.filter_by(relation_id=relation_id).one()
        assert model.reverse


def test_delete(app):
    with app.app_context():
        count = ProductRelation.query.count()
        model = ProductRelation.query.first()
        relation_id = model.relation_id

        ops.delete_product_relation(relation_id=relation_id)
        ops.commit()

    with app.app_context():
        assert ProductRelation.query.count() == (count - 2)
        model = ProductRelation.query.filter_by(
            relation_id=relation_id
        ).one_or_none()
        assert model is None



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_simple 1'] = [
    (
        (
        ),
        {
        }
    ),
    (
        (
            'https://api.github.com/graphql'
        ,),
        {
            'headers': {
                'Authorization': 'token token-xxx'
            },
            'json': {
                'query': '{ viewer { login name avatarUrl } }'
            }
        }
    )
]

snapshots['test_variables 1'] = [
    (
        (
        ),
        {
        }
    ),
    (
        (
            'https://api.github.com/graphql'
        ,),
        {
            'headers': {
                'Authorization': 'token token-xxx'
            },
            'json': {
                'query': '''
      query Organization($org_login: String!) {
        organization(login: $org_login) {
          login
          avatarUrl
        }
      }
    ''',
                'variables': {
                    'org_login': 'urban-octo-disco'
                }
            }
        }
    )
]



import graphene
from graphene_sqlalchemy import SQLAlchemyObjectType

from ...models import WebConfig as WebConfigModel

from ..connection import CountedConnection


##__________________________________________________________________||
class WebConfig(SQLAlchemyObjectType):
    """Web configuration"""

    class Meta:
        model = WebConfigModel
        interfaces = (graphene.relay.Node,)
        connection_class = CountedConnection


def resolve_web_config(parent, info, **kwargs):
    return WebConfig.get_query(info).one_or_none()


web_config_field = graphene.Field(WebConfig, resolver=resolve_web_config)

##__________________________________________________________________||



from ...db.sa import sa


##__________________________________________________________________||
class TypeFieldAssociation(sa.Model):
    __tablename__ = "type_field_association"
    type_id = sa.Column(
        #
        sa.Integer(),
        sa.ForeignKey("product_types.type_id"),
        primary_key=True,
    )
    field_id = sa.Column(
        #
        sa.Integer(),
        sa.ForeignKey("field.field_id"),
        primary_key=True,
    )
    type_ = sa.relationship(
        "ProductType",
        backref=sa.backref("fields", cascade="all, delete-orphan"),
    )
    field = sa.relationship(
        "Field",
        backref=sa.backref("entry_types"),
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.type_name!r} {self.field_name!r}>"

    @property
    def type_name(self):
        # used in __repr__()
        try:
            return self.type_.name
        except BaseException:
            return self.type_

    @property
    def field_name(self):
        # used in __repr__()
        try:
            return self.field.name
        except BaseException:
            return self.field


##__________________________________________________________________||



"""add fields to products

Revision ID: babb78f6f6b3
Revises: c090bba0e451
Create Date: 2021-02-25 13:30:06.303180

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'babb78f6f6b3'
down_revision = 'c090bba0e451'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.add_column(sa.Column('posting_git_hub_user_id', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('updating_git_hub_user_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_products_posting_git_hub_user_id_github_users'), 'github_users', ['posting_git_hub_user_id'], ['user_id'])
        batch_op.create_foreign_key(batch_op.f('fk_products_updating_git_hub_user_id_github_users'), 'github_users', ['updating_git_hub_user_id'], ['user_id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_products_updating_git_hub_user_id_github_users'), type_='foreignkey')
        batch_op.drop_constraint(batch_op.f('fk_products_posting_git_hub_user_id_github_users'), type_='foreignkey')
        batch_op.drop_column('updating_git_hub_user_id')
        batch_op.drop_column('posting_git_hub_user_id')

    # ### end Alembic commands ###



from ...db.sa import sa


##__________________________________________________________________||
class ProductFilePath(sa.Model):
    __tablename__ = "product_file_paths"
    path_id = sa.Column(sa.Integer(), primary_key=True)
    path = sa.Column(sa.Text())
    note = sa.Column(sa.Text())
    product_id = sa.Column(sa.ForeignKey("products.product_id"))
    product = sa.relationship(
        "Product", backref=sa.backref("paths", cascade="all, delete-orphan")
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.path_shorten!r}>"

    @property
    def path_shorten(self):
        try:
            placeholder = "..."
            width = max(20, len(placeholder))
            if len(self.path) <= width:
                return self.path
            return placeholder + self.path[-(width - len(placeholder)) :]
        except BaseException:
            return self.path


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_error 1'] = [
    (
        (
        ),
        {
        }
    ),
    (
        (
            'https://github.com/login/oauth/access_token'
        ,),
        {
            'headers': {
                'Accept': 'application/vnd.github.v3+json, application/json'
            },
            'json': {
                'client_id': 'client_id_0123',
                'client_secret': 'client_secret_0123',
                'code': 'code-xyz',
                'grant_type': 'authorization_code',
                'redirect_uri': 'http://localhost/auth'
            }
        }
    )
]

snapshots['test_success 1'] = [
    (
        (
        ),
        {
        }
    ),
    (
        (
            'https://github.com/login/oauth/access_token'
        ,),
        {
            'headers': {
                'Accept': 'application/vnd.github.v3+json, application/json'
            },
            'json': {
                'client_id': 'client_id_0123',
                'client_secret': 'client_secret_0123',
                'code': 'code-xyz',
                'grant_type': 'authorization_code',
                'redirect_uri': 'http://localhost/auth'
            }
        }
    )
]



import graphene

from .. import type_

from .... import ops


##__________________________________________________________________||
class CommonInputFields:
    name = graphene.String(
        required=True,
        description="The name of the field",
    )


class CreateFieldInput(graphene.InputObjectType, CommonInputFields):
    """Input to createField()"""
    type_ = type_.FieldType()


class UpdateFieldInput(graphene.InputObjectType, CommonInputFields):
    """Input to updateField()"""


##__________________________________________________________________||
class CreateField(graphene.Mutation):
    """Create a field"""

    class Arguments:
        input = CreateFieldInput(required=True)

    ok = graphene.Boolean()
    field = graphene.Field(lambda: type_.Field)

    def mutate(root, info, input):
        model = ops.create_field(**input)
        ops.commit()
        ok = True
        return CreateField(field=model, ok=ok)


class UpdateField(graphene.Mutation):
    """Update a field"""

    class Arguments:
        field_id = graphene.Int(required=True)
        input = UpdateFieldInput(required=True)

    ok = graphene.Boolean()
    field = graphene.Field(lambda: type_.Field)

    def mutate(root, info, field_id, input):
        name = input.pop('name')
        model = ops.update_field(field_id, name)
        ops.commit()
        ok = True
        return UpdateField(field=model, ok=ok)


class DeleteField(graphene.Mutation):
    """Delete a field"""

    class Arguments:
        field_id = graphene.Int(description="The fieldId of the field")

    ok = graphene.Boolean()

    def mutate(root, info, field_id):
        ops.delete_field(field_id)
        ops.commit()
        ok = True
        return DeleteField(ok=ok)


##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import Product, ProductRelation


##__________________________________________________________________||
def test_cascade_delete_children(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # delete child1
    with app.app_context():
        child1 = Product.query.filter_by(name="child1").one_or_none()
        sa.session.delete(child1)
        sa.session.commit()

    #
    #
    #
    #  +---------+
    #  |         |
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # assert
    with app.app_context():

        parent1 = Product.query.filter_by(name="parent1").one_or_none()
        child1 = Product.query.filter_by(name="child1").one_or_none()
        child2 = Product.query.filter_by(name="child2").one_or_none()

        assert parent1 is not None
        assert child1 is None
        assert child2 is not None

        relations = ProductRelation.query.all()
        assert 2 == len(relations)

        assert 1 == len(parent1.relations)
        assert child2 == parent1.relations[0].other
        assert parent1 == child2.relations[0].other
        assert parent1.relations[0].reverse == child2.relations[0]

    # delete child2
    with app.app_context():
        child2 = Product.query.filter_by(name="child2").one_or_none()
        sa.session.delete(child2)
        sa.session.commit()

    #
    #  +---------+
    #  |         |
    #  | parent1 |
    #  |         |
    #  +---------+
    #

    # assert
    with app.app_context():

        parent1 = Product.query.filter_by(name="parent1").one_or_none()
        child1 = Product.query.filter_by(name="child1").one_or_none()
        child2 = Product.query.filter_by(name="child2").one_or_none()

        assert parent1 is not None
        assert child1 is None
        assert child2 is None

        relations = ProductRelation.query.all()
        assert 0 == len(relations)

        assert 0 == len(parent1.relations)


##__________________________________________________________________||
def test_cascade_delete_parent(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # delete parent1
    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").first()
        sa.session.delete(parent1)
        sa.session.commit()

    #                              +--------+
    #                              |        |
    #                              | child1 |
    #                              |        |
    #                              +--------+
    #
    #                              +--------+
    #                              |        |
    #                              | child2 |
    #                              |        |
    #                              +--------+

    # assert
    with app.app_context():

        parent1 = Product.query.filter_by(name="parent1").one_or_none()
        child1 = Product.query.filter_by(name="child1").one_or_none()
        child2 = Product.query.filter_by(name="child2").one_or_none()

        assert parent1 is None
        assert child1 is not None
        assert child2 is not None

        relations = ProductRelation.query.all()
        assert 0 == len(relations)


##__________________________________________________________________||
def test_cascade_delete_relations(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # delete the relation from child1 to parent1
    with app.app_context():
        child1 = Product.query.filter_by(name="child1").one_or_none()
        sa.session.delete(child1.relations[0])
        sa.session.commit()

    #                              +--------+
    #                              |        |
    #                              | child1 |
    #  +---------+                 |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # assert
    with app.app_context():

        parent1 = Product.query.filter_by(name="parent1").one_or_none()
        child1 = Product.query.filter_by(name="child1").one_or_none()
        child2 = Product.query.filter_by(name="child2").one_or_none()

        assert parent1 is not None
        assert child1 is not None
        assert child2 is not None

        relations = ProductRelation.query.all()
        assert 2 == len(relations)

        assert 1 == len(parent1.relations)
        assert child2 == parent1.relations[0].other
        assert parent1 == child2.relations[0].other
        assert parent1.relations[0].reverse == child2.relations[0]

    # delete the relation from parent1 to child2
    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").one_or_none()
        sa.session.delete(parent1.relations[0])
        sa.session.commit()

    #                              +--------+
    #                              |        |
    #                              | child1 |
    #  +---------+                 |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+                 |        |
    #                              | child2 |
    #                              |        |
    #                              +--------+

    # assert
    with app.app_context():

        parent1 = Product.query.filter_by(name="parent1").one_or_none()
        child1 = Product.query.filter_by(name="child1").one_or_none()
        child2 = Product.query.filter_by(name="child2").one_or_none()

        assert parent1 is not None
        assert child1 is not None
        assert child2 is not None

        relations = ProductRelation.query.all()
        assert 0 == len(relations)


##__________________________________________________________________||



"""change column type

Revision ID: a4105346b478
Revises: e077a31ad00b
Create Date: 2021-09-22 14:57:25.293805

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "a4105346b478"
down_revision = "e077a31ad00b"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("attribute_text", schema=None) as batch_op:
        batch_op.alter_column(
            "value", type_=sa.UnicodeText(), existing_type=sa.Text()
        )


def downgrade():
    pass



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[query] 1'] = {
    'data': {
        'webConfig': {
            'configId': '1',
            'devtoolLoadingstate': True,
            'headTitle': 'Head Title',
            'productCreationDialog': False,
            'productDeletionDialog': True,
            'productUpdateDialog': True,
            'toolbarTitle': 'Toolbar Title'
        }
    }
}



"""add columns to product_types

Revision ID: d7cac99bbffa
Revises: cc64c8ef2115
Create Date: 2020-05-22 11:58:47.182210

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd7cac99bbffa'
down_revision = 'cc64c8ef2115'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('product_types', sa.Column('icon', sa.Text(), nullable=True))
    op.add_column('product_types', sa.Column('indef_article', sa.Text(), nullable=True))
    op.add_column('product_types', sa.Column('plural', sa.Text(), nullable=True))
    op.add_column('product_types', sa.Column('singular', sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('product_types', 'singular')
    op.drop_column('product_types', 'plural')
    op.drop_column('product_types', 'indef_article')
    op.drop_column('product_types', 'icon')
    # ### end Alembic commands ###



from acondbs.db.sa import sa
from acondbs.models import GitHubUser


##__________________________________________________________________||
def test_entry(app_empty):
    app = app_empty

    user1 = GitHubUser(
        login="octocat",
        git_hub_id="04:User583231",
        name="The Octocat",
        avatar_url="https://avatars3.githubusercontent.com/u/583231?v=4",
    )

    with app.app_context():
        sa.session.add(user1)
        sa.session.commit()

    with app.app_context():
        user1 = GitHubUser.query.filter_by(login="octocat").one()
        assert "octocat" == user1.login


##__________________________________________________________________||



import graphene

from .. import type_

from .... import ops


##__________________________________________________________________||
class CommonInputFields:
    name = graphene.String(
        required=True, description="The name of the product type"
    )
    order = graphene.Int(
        description=(
            "The order in which the type is displayed, for example, "
            "in navigation bars."
        )
    )
    indef_article = graphene.String(
        description=(
            'The indefinite article placed before the singular noun "'
            'i.e., "a" or "an". '
        )
    )
    singular = graphene.String(
        description=("The singular noun, the product type name in singular.")
    )
    plural = graphene.String(
        description=("The plural noun, the product type name in plural.")
    )
    icon = graphene.String(
        description=(
            "A name of the icon from https://materialdesignicons.com/"
        )
    )
    field_ids = graphene.List(
        graphene.Int,
        description=("The field IDs")
    )


class CreateProductTypeInput(graphene.InputObjectType, CommonInputFields):
    """Input to createProductType()"""


class UpdateProductTypeInput(graphene.InputObjectType, CommonInputFields):
    """Input to updateProductType()"""


##__________________________________________________________________||
class CreateProductType(graphene.Mutation):
    """Create a product type"""

    class Arguments:
        input = CreateProductTypeInput(required=True)

    ok = graphene.Boolean()
    product_type = graphene.Field(lambda: type_.ProductType)

    def mutate(root, info, input):
        model = ops.create_product_type(**input)
        ops.commit()
        ok = True
        return CreateProductType(product_type=model, ok=ok)


class UpdateProductType(graphene.Mutation):
    """Update a product type"""

    class Arguments:
        type_id = graphene.Int(required=True)
        input = UpdateProductTypeInput(required=True)

    ok = graphene.Boolean()
    product_type = graphene.Field(lambda: type_.ProductType)

    def mutate(root, info, type_id, input):
        model = ops.update_product_type(type_id, **input)
        ops.commit()
        ok = True
        return UpdateProductType(product_type=model, ok=ok)


class DeleteProductType(graphene.Mutation):
    """Delete a product type"""

    class Arguments:
        type_id = graphene.Int(description="The typeId of the product type")

    ok = graphene.Boolean()

    def mutate(root, info, type_id):
        ops.delete_product_type(type_id)
        ops.commit()
        ok = True
        return DeleteProductType(ok=ok)


##__________________________________________________________________||



import pytest

from ...funcs import assert_query

from ..gql import QUERY_PRODUCT_TYPE

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


QUERY_PRODUCT_TYPE_SORT_PRODUCTS = """
{
  productType(typeId: 1) {
    typeId
    name
    order
    indefArticle
    singular
    plural
    icon
    products(sort: DATE_PRODUCED_DESC) {
      edges {
        node {
          name
        }
      }
    }
  }
}
"""


##__________________________________________________________________||
params = [
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_TYPE,
            "variables": {"typeId": 1},
        },
        id="type_id",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_TYPE,
            "variables": {"name": "map"},
        },
        id="name",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_TYPE,
            "variables": {"typeId": 1, "name": "map"},
        },
        id="type_id-and-name",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_TYPE,
            "variables": {"typeId": 2, "name": "map"},
        },
        id="type_id-and-name-nonexistent",
    ),
    pytest.param(
        {
            #
            "query": QUERY_PRODUCT_TYPE_SORT_PRODUCTS,
        },
        id="type_id-sort-products",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)


##__________________________________________________________________||



import pytest

from ...funcs import assert_query

GITHUB_USER_FRAGMENT = '''
fragment GitHubUserFragment on GitHubUser {
  login
  name
  avatarUrl
  memberships {
    edges {
      node {
        org {
          login
        }
      }
    }
  }
}
'''

ALL_GITHUB_USERS = '''
{
  allGitHubUsers {
    totalCount
    edges {
      node {
        ...GitHubUserFragment
      }
    }
  }
}
''' + GITHUB_USER_FRAGMENT

ALL_GITHUB_USERS_WITH_FILTER = '''
query AllGitHubUsers($orgMember: Boolean){
  allGitHubUsers(filters: { orgMember: $orgMember }) {
    totalCount
    edges {
      node {
        ...GitHubUserFragment
      }
    }
  }
}
''' + GITHUB_USER_FRAGMENT

HEADERS = {
    "Authorization": "Bearer token1"  # user1
}

##__________________________________________________________________||
params = [
    pytest.param(
        {"query": ALL_GITHUB_USERS},
        id='one'
    ),
    pytest.param(
        {"query": ALL_GITHUB_USERS_WITH_FILTER, 'variables': {'orgMember': True}},
        id='filter-on'
    ),
    pytest.param(
        {"query": ALL_GITHUB_USERS_WITH_FILTER, 'variables': {'orgMember': False}},
        id='filter-off'
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)

##__________________________________________________________________||



from ...db.sa import sa


##__________________________________________________________________||
class WebConfig(sa.Model):
    __tablename__ = "web_config"
    config_id = sa.Column(sa.Integer(), primary_key=True)
    head_title = sa.Column(sa.Text())
    toolbar_title = sa.Column(sa.Text())
    devtool_loadingstate = sa.Column(sa.Boolean())
    product_creation_dialog = sa.Column(sa.Boolean())
    product_update_dialog = sa.Column(sa.Boolean())
    product_deletion_dialog = sa.Column(sa.Boolean())


##__________________________________________________________________||



import pytest

from ....funcs import assert_mutation

from ...gql import MUTATION_DELETE_PRODUCT

QEURY = """
{
  allProducts {
    edges {
      node {
        name
      }
    }
  }
  allProductRelations {
    edges {
      node {
        type_ {
          name
        }
        self_ {
          name
        }
        other {
          name
        }
      }
    }
  }
  allProductFilePaths {
    edges {
      node {
        path
      }
    }
  }
}
"""


HEADERS = {"Authorization": "Bearer 39d86487d76a84087f1da599c872dac4473e5f07"}  # user1


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": MUTATION_DELETE_PRODUCT, "variables": {"productId": 1}},
        {"query": QEURY},
        id="delete",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_success(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):
    success = True
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": MUTATION_DELETE_PRODUCT, "variables": {"productId": 512}},
        {"query": QEURY},
        id="error",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_error(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):
    success = False
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



import pytest

from itertools import permutations

from acondbs.db.sa import sa
from acondbs.models import (
    ProductType,
    Product,
    ProductRelation,
    ProductRelationType,
)

##__________________________________________________________________||
params = list(permutations([1, 2, 3]))
# i.e., [(1, 2, 3), (1, 3, 2), (2, 1, 3), (2, 3, 1), (3, 1, 2), (3, 2, 1)]

ids = ["-".join(["{}".format(e) for e in p]) for p in params]
# i.e., ['1-2-3', '1-3-2', '2-1-3', '2-3-1', '3-1-2', '3-2-1']


@pytest.mark.parametrize("perm", params, ids=ids)
def test_permutations(app_empty, perm):

    app = app_empty

    #
    # +---------+                 +--------+
    # |         |  --(child)-->   |        |
    # | parent1 |       |         | child1 |
    # |         |  <-(parent)--   |        |
    # +---------+                 +--------+
    #

    type_ = ProductType(name="robot")
    parent1 = Product(name="parent1", type_=type_)
    child1 = Product(name="child1", type_=type_)

    relation_type_parent = ProductRelationType(name="parent")
    relation_type_child = ProductRelationType(name="child")
    relation_type_parent.reverse = relation_type_child
    relation_type_child.reverse = relation_type_parent

    relation_parent1_to_child1 = ProductRelation()

    # set 'type_', 'self_', 'other' in different orders
    # each triggers an event in which a reverse is automatically set
    for p in perm:
        if p == 1:
            relation_parent1_to_child1.type_ = relation_type_child
        elif p == 2:
            relation_parent1_to_child1.self_ = parent1
        elif p == 3:
            relation_parent1_to_child1.other = child1

    with app.app_context():
        sa.session.add(parent1)
        sa.session.commit()

    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").first()
        child1 = Product.query.filter_by(name="child1").first()

        assert 1 == len(parent1.relations)
        assert 1 == len(child1.relations)

        assert "child" == parent1.relations[0].type_.name
        assert "parent" == child1.relations[0].type_.name

        assert child1 is parent1.relations[0].other

        assert parent1 is child1.relations[0].other

        assert parent1.relations[0] is child1.relations[0].reverse
        assert parent1.relations[0].reverse is child1.relations[0]


##__________________________________________________________________||
@pytest.mark.parametrize("perm", params, ids=ids)
def test_self_reverse_type(app_empty, perm):

    app = app_empty

    #
    # +----------+                 +----------+
    # |          |  --(sibling)--> |          |
    # | sibling1 |       |         | sibling2 |
    # |          |  <-(sibling)--  |          |
    # +----------+                 +----------+
    #

    type_ = ProductType(name="robot")
    sibling1 = Product(name="sibling1", type_=type_)
    sibling2 = Product(name="sibling2", type_=type_)

    relation_type_sibling = ProductRelationType(name="sibling")
    relation_type_sibling.reverse = relation_type_sibling

    relation_sibling1_to_sibling2 = ProductRelation()

    # set 'type_', 'self_', 'other' in different orders
    # each triggers an event in which a reverse is automatically set
    for p in perm:
        if p == 1:
            relation_sibling1_to_sibling2.type_ = relation_type_sibling
        elif p == 2:
            relation_sibling1_to_sibling2.self_ = sibling1
        elif p == 3:
            relation_sibling1_to_sibling2.other = sibling2

    with app.app_context():
        sa.session.add(sibling1)
        sa.session.commit()

    with app.app_context():
        sibling1 = Product.query.filter_by(name="sibling1").first()
        sibling2 = Product.query.filter_by(name="sibling2").first()

        assert 1 == len(sibling1.relations)
        assert 1 == len(sibling2.relations)

        assert "sibling" == sibling1.relations[0].type_.name
        assert "sibling" == sibling2.relations[0].type_.name

        assert sibling1.relations[0].type_ is sibling2.relations[0].type_

        assert sibling2 is sibling1.relations[0].other

        assert sibling1 is sibling2.relations[0].other

        assert sibling1.relations[0] is sibling2.relations[0].reverse
        assert sibling1.relations[0].reverse is sibling2.relations[0]


##__________________________________________________________________||
@pytest.mark.xfail(reason="events won't be triggered by foreign keys")
def test_id(app_empty):

    # This test doesn't pass because the "set" events to the
    # relationships won't be triggered when the foreign keys are set.

    # It might be possible to pass this test with before_flush().
    # Useful links about before_flush():
    # https://docs.sqlalchemy.org/en/14/orm/session_events.html#persistence-events
    # https://stackoverflow.com/questions/36731020/sqlalchemy-orm-event-hook-for-attribute-persisted

    app = app_empty

    #
    # +---------+                 +--------+
    # |         |  --(child)-->   |        |
    # | parent1 |       |         | child1 |
    # |         |  <-(parent)--   |        |
    # +---------+                 +--------+
    #

    type_ = ProductType(name="robot")
    parent1 = Product(product_id=1, name="parent1", type_=type_)
    child1 = Product(product_id=2, name="child1", type_=type_)

    relation_type_parent = ProductRelationType(type_id=1, name="parent")
    relation_type_child = ProductRelationType(type_id=2, name="child")
    relation_type_parent.reverse = relation_type_child

    relation_parent1_to_child1 = ProductRelation(
        type_id=2, self_product_id=1, other_product_id=2
    )

    with app.app_context():
        sa.session.add(parent1)
        sa.session.add(child1)
        sa.session.add(relation_type_parent)
        sa.session.add(relation_parent1_to_child1)
        sa.session.flush()
        sa.session.commit()

    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").first()
        child1 = Product.query.filter_by(name="child1").first()

        assert 1 == len(parent1.relations)
        assert 1 == len(child1.relations)

        assert "child" == parent1.relations[0].type_.name
        assert "parent" == child1.relations[0].type_.name

        assert child1 is parent1.relations[0].other

        assert parent1 is child1.relations[0].other

        assert parent1.relations[0] is child1.relations[0].reverse
        assert parent1.relations[0].reverse is child1.relations[0]


##__________________________________________________________________||
def test_attach_to_self(app_empty):

    app = app_empty

    #
    # +---------+                 +--------+
    # |         |  --(child)-->   |        |
    # | parent1 |       |         | child1 |
    # |         |  <-(parent)--   |        |
    # +---------+                 +--------+
    #

    type_ = ProductType(name="robot")
    parent1 = Product(product_id=1, name="parent1", type_=type_)
    child1 = Product(product_id=2, name="child1", type_=type_)

    relation_type_parent = ProductRelationType(type_id=1, name="parent")
    relation_type_child = ProductRelationType(type_id=2, name="child")
    relation_type_parent.reverse = relation_type_child

    relation_parent1_to_child1 = ProductRelation(
        type_=relation_type_child, other=child1
    )

    parent1.relations = [relation_parent1_to_child1]

    with app.app_context():
        sa.session.add(parent1)
        sa.session.add(child1)
        sa.session.add(relation_type_parent)
        sa.session.add(relation_parent1_to_child1)
        sa.session.flush()
        sa.session.commit()

    with app.app_context():
        parent1 = Product.query.filter_by(name="parent1").first()
        child1 = Product.query.filter_by(name="child1").first()

        assert 1 == len(parent1.relations)
        assert 1 == len(child1.relations)

        assert "child" == parent1.relations[0].type_.name
        assert "parent" == child1.relations[0].type_.name

        assert child1 is parent1.relations[0].other

        assert parent1 is child1.relations[0].other

        assert parent1.relations[0] is child1.relations[0].reverse
        assert parent1.relations[0].reverse is child1.relations[0]


##__________________________________________________________________||



from .fragment_product_relation_type import FRAGMENT_PRODUCT_RELATION_TYPE

FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION = '''
fragment fragmentProductRelationTypeConnection on ProductRelationTypeConnection {
  edges {
    node {
      ...fragmentProductRelationType
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE



from ..fragments import FRAGMENT_PRODUCT_RELATION

QUERY_PRODUCT_RELATION = """
query ProductRelation($relationId: Int) {
  productRelation(relationId: $relationId) {
    ...fragmentProductRelation
  }
}
""" + FRAGMENT_PRODUCT_RELATION



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables
from acondbs.db.sa import sa
from acondbs.models import (
    GitHubOrg,
    GitHubUser,
    GitHubOrgMembership,
    GitHubToken,
    AccountAdmin,
)


##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri = "sqlite:///:memory:"
    y = create_app(
        SQLALCHEMY_DATABASE_URI=database_uri,
        GITHUB_AUTH_AUTHORIZE_URL="https://github.com/login/oauth/authorize",
        GITHUB_AUTH_TOKEN_URL="https://github.com/login/oauth/access_token",
        GITHUB_AUTH_CLIENT_ID="client_id_0123456789",
        GITHUB_AUTH_CLIENT_SECRET="client_secret_abcdefghijklmnupqrstuvwxyz",
        GITHUB_AUTH_REDIRECT_URI="http://localhost:8080/signin",
    )
    with y.app_context():
        define_tables()
    yield y


@pytest.fixture
def app(app_empty):

    y = app_empty

    #
    #  +------+
    #  |      |       +-------+       +--------+
    #  | org1 | ----- |       | --+-- | token1 |
    #  |      |       | user1 |   |   +--------+
    #  +------+   +-- |       |   |
    #             |   +-------+   |   +--------+
    #             |               +-- | token2 |
    #  +------+   |                   +--------+
    #  |      | --+   +-------+
    #  | org2 |       |       |       +--------+
    #  |      | ----- | user2 | ----- | token3 |
    #  +------+       |       |       +--------+
    #                 +-------+
    #
    #  +------+       +-------+
    #  |      |       |       |       +--------+
    #  | org3 |       | user3 | ----- | token4 |
    #  |      |       |       |       +--------+
    #  +------+       +-------+
    #
    #

    org1 = GitHubOrg(login="org1", git_hub_id="012:Organization1")
    org2 = GitHubOrg(login="org2", git_hub_id="012:Organization2")
    org3 = GitHubOrg(login="org3", git_hub_id="012:Organization3")

    user1 = GitHubUser(
        user_id=1,
        login="user1",
        git_hub_id="04:User1",
        name="User One",
        avatar_url="avatar.com/user1",
    )
    user2 = GitHubUser(
        user_id=2,
        login="user2",
        git_hub_id="04:User2",
        name="User Two",
        avatar_url="avatar.com/user2",
    )
    user3 = GitHubUser(
        user_id=3,
        login="user3",
        git_hub_id="04:User3",
        name="User Three",
        avatar_url="avatar.com/user3",
    )

    GitHubToken(token_id=1, token="token1", scope="read:org", user=user1)
    GitHubToken(token_id=2, token="token2", scope="", user=user1)
    GitHubToken(token_id=3, token="token3", scope="", user=user2)
    GitHubToken(token_id=4, token="token4", scope="", user=user3)

    GitHubOrgMembership(org=org1, member=user1)
    GitHubOrgMembership(org=org2, member=user1)
    GitHubOrgMembership(org=org2, member=user2)

    admin1 = AccountAdmin(git_hub_login="user1")
    admin2 = AccountAdmin(git_hub_login="user3")

    with y.app_context():
        sa.session.add(org1)
        sa.session.add(org2)
        sa.session.add(org3)
        sa.session.add(user1)
        sa.session.add(user2)
        sa.session.add(user3)
        sa.session.add(admin1)
        sa.session.add(admin2)
        sa.session.commit()
    yield y


##__________________________________________________________________||



"""replace date with data time in products

Revision ID: c090bba0e451
Revises: 4c32365bb655
Create Date: 2021-02-25 09:39:10.272222

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c090bba0e451'
down_revision = '4c32365bb655'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('account_admins', schema=None) as batch_op:
        batch_op.drop_constraint('uq_account_admins_github_login', type_='unique')
        batch_op.create_unique_constraint(batch_op.f('uq_account_admins_git_hub_login'), ['git_hub_login'])

    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.add_column(sa.Column('time_posted', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('time_updated', sa.DateTime(), nullable=True))
        batch_op.drop_column('date_updated')
        batch_op.drop_column('date_posted')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.add_column(sa.Column('date_posted', sa.DATE(), nullable=True))
        batch_op.add_column(sa.Column('date_updated', sa.DATE(), nullable=True))
        batch_op.drop_column('time_updated')
        batch_op.drop_column('time_posted')

    with op.batch_alter_table('account_admins', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('uq_account_admins_git_hub_login'), type_='unique')
        batch_op.create_unique_constraint('uq_account_admins_github_login', ['git_hub_login'])

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[filter] 1'] = {
    'data': {
        'allGitHubTokens': {
            'edges': [
                {
                    'node': {
                        'scope': 'read:org',
                        'timeCreated': '2021-01-04T14:32:20',
                        'tokenId': '1',
                        'tokenMasked': 'XXXXXXXXXXXXXXX',
                        'user': {
                            'login': 'user1'
                        }
                    }
                }
            ],
            'totalCount': 1
        }
    }
}

snapshots['test_schema[simple] 1'] = {
    'data': {
        'allGitHubTokens': {
            'edges': [
                {
                    'node': {
                        'scope': 'read:org',
                        'timeCreated': '2021-01-04T14:32:20',
                        'tokenId': '1',
                        'tokenMasked': 'XXXXXXXXXXXXXXX',
                        'user': {
                            'login': 'user1'
                        }
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'timeCreated': '2021-01-04T14:32:20',
                        'tokenId': '2',
                        'tokenMasked': 'XXXXXXXXXXXXXXX',
                        'user': {
                            'login': 'user1'
                        }
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'timeCreated': '2021-01-04T14:32:20',
                        'tokenId': '3',
                        'tokenMasked': 'XXXXXXXXXXXXXXX',
                        'user': {
                            'login': 'user2'
                        }
                    }
                },
                {
                    'node': {
                        'scope': '',
                        'timeCreated': '2021-01-04T14:32:20',
                        'tokenId': '4',
                        'tokenMasked': 'XXXXXXXXXXXXXXX',
                        'user': {
                            'login': 'user3'
                        }
                    }
                }
            ],
            'totalCount': 4
        }
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[error] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[delete] 1'] = {
    'data': {
        'deleteProduct': {
            'ok': True
        }
    }
}

snapshots['test_schema_success[delete] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}



import pytest

from ...funcs import assert_query

ALL_GITHUB_ORGS = '''
{
  allGitHubOrgs {
    totalCount
    edges {
      node {
        login
        memberships {
          totalCount
          edges {
            node {
              member {
                login
              }
            }
          }
        }
      }
    }
  }
}
'''

HEADERS = {
    "Authorization": "Bearer token1"  # user1
}


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": ALL_GITHUB_ORGS},
        id="one",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)

##__________________________________________________________________||



import pytest

from flask_sqlalchemy import BaseQuery

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.


##__________________________________________________________________||
@pytest.fixture
def app(app):
    y = app

    type_map = ProductType(name="map")
    map1 = Product(name="map1", type_=type_map)  # noqa: F841
    map2 = Product(name="map2", type_=type_map)  # noqa: F841
    map3 = Product(name="map3", type_=type_map)  # noqa: F841

    with y.app_context():
        sa.session.add(type_map)
        sa.session.commit()

    yield y


##__________________________________________________________________||
def test_context(app):

    # query cannot be accessed outside of the app context
    with pytest.raises(RuntimeError):
        Product.query

    with app.app_context():
        Product.query


def test_query_all(app):

    with app.app_context():

        # query is an instance of BaseQuery
        query = Product.query
        assert isinstance(query, BaseQuery)

        # query.all() returns a list of products
        results = query.all()
        # e.g., [<Product 1>, <Product 2>, <Product 3>]

        assert isinstance(results[0], Product)


def test_query_filter(app):

    with app.app_context():

        # filter_by() returns an instance of BaseQuery
        query = Product.query.filter_by(name="map1")
        assert isinstance(query, BaseQuery)

        # the results are a list with one element
        results = query.all()
        assert 1 == len(results)
        assert ["map1"] == [e.name for e in results]

        # first() returns a product
        product = query.first()
        assert isinstance(product, Product)
        assert "map1" == product.name


def test_query_filter_nonexistent(app):

    with app.app_context():

        query = Product.query.filter_by(name="no-such-product")
        assert isinstance(query, BaseQuery)

        # the results are an empty list
        results = query.all()
        assert [] == results

        # first() returns None
        product = query.first()
        assert product is None


##__________________________________________________________________||



"""add a field "time_created" to github_tokens

Revision ID: 2b15f740eed9
Revises: 936fc5b02f16
Create Date: 2021-01-04 12:03:13.963073

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2b15f740eed9'
down_revision = '936fc5b02f16'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_tokens', schema=None) as batch_op:
        batch_op.add_column(sa.Column('time_created', sa.DateTime(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('github_tokens', schema=None) as batch_op:
        batch_op.drop_column('time_created')

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[error-nonexistent] 1'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-spotlight-beam',
                        'indefArticle': 'a',
                        'name': 'beam',
                        'order': 1,
                        'plural': 'beams',
                        'products': {
                            'edges': [
                            ]
                        },
                        'singular': 'beam',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'map'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'map'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'type_': {
                                            'name': 'map'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-map',
                        'indefArticle': 'a',
                        'name': 'map',
                        'order': 2,
                        'plural': 'maps',
                        'products': {
                            'edges': [
                                {
                                    'node': {
                                        'name': 'map1'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map2'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map3'
                                    }
                                }
                            ]
                        },
                        'singular': 'map',
                        'typeId': '1'
                    }
                }
            ],
            'totalCount': 2
        }
    }
}

snapshots['test_schema_success[empty-fields] 1'] = {
    'data': {
        'updateProductType': {
            'ok': True,
            'productType': {
                'fields': {
                    'edges': [
                    ]
                },
                'icon': 'mdi-compass',
                'indefArticle': 'a',
                'name': 'compass',
                'order': 5,
                'plural': 'compasses',
                'products': {
                    'edges': [
                        {
                            'node': {
                                'name': 'map1'
                            }
                        },
                        {
                            'node': {
                                'name': 'map2'
                            }
                        },
                        {
                            'node': {
                                'name': 'map3'
                            }
                        }
                    ]
                },
                'singular': 'compass',
                'typeId': '1'
            }
        }
    }
}

snapshots['test_schema_success[empty-fields] 2'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-spotlight-beam',
                        'indefArticle': 'a',
                        'name': 'beam',
                        'order': 1,
                        'plural': 'beams',
                        'products': {
                            'edges': [
                            ]
                        },
                        'singular': 'beam',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'fields': {
                            'edges': [
                            ]
                        },
                        'icon': 'mdi-compass',
                        'indefArticle': 'a',
                        'name': 'compass',
                        'order': 5,
                        'plural': 'compasses',
                        'products': {
                            'edges': [
                                {
                                    'node': {
                                        'name': 'map1'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map2'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map3'
                                    }
                                }
                            ]
                        },
                        'singular': 'compass',
                        'typeId': '1'
                    }
                }
            ],
            'totalCount': 2
        }
    }
}

snapshots['test_schema_success[fields-unchanged] 1'] = {
    'data': {
        'updateProductType': {
            'ok': True,
            'productType': {
                'fields': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'type_': {
                                    'name': 'compass'
                                }
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'type_': {
                                    'name': 'compass'
                                }
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                },
                                'type_': {
                                    'name': 'compass'
                                }
                            }
                        }
                    ]
                },
                'icon': 'mdi-compass',
                'indefArticle': 'a',
                'name': 'compass',
                'order': 5,
                'plural': 'compasses',
                'products': {
                    'edges': [
                        {
                            'node': {
                                'name': 'map1'
                            }
                        },
                        {
                            'node': {
                                'name': 'map2'
                            }
                        },
                        {
                            'node': {
                                'name': 'map3'
                            }
                        }
                    ]
                },
                'singular': 'compass',
                'typeId': '1'
            }
        }
    }
}

snapshots['test_schema_success[fields-unchanged] 2'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-spotlight-beam',
                        'indefArticle': 'a',
                        'name': 'beam',
                        'order': 1,
                        'plural': 'beams',
                        'products': {
                            'edges': [
                            ]
                        },
                        'singular': 'beam',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'compass'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'compass'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'type_': {
                                            'name': 'compass'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-compass',
                        'indefArticle': 'a',
                        'name': 'compass',
                        'order': 5,
                        'plural': 'compasses',
                        'products': {
                            'edges': [
                                {
                                    'node': {
                                        'name': 'map1'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map2'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map3'
                                    }
                                }
                            ]
                        },
                        'singular': 'compass',
                        'typeId': '1'
                    }
                }
            ],
            'totalCount': 2
        }
    }
}

snapshots['test_schema_success[update] 1'] = {
    'data': {
        'updateProductType': {
            'ok': True,
            'productType': {
                'fields': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'type_': {
                                    'name': 'compass'
                                }
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'field_four',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'type_': {
                                    'name': 'compass'
                                }
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'field_five',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'type_': {
                                    'name': 'compass'
                                }
                            }
                        }
                    ]
                },
                'icon': 'mdi-compass',
                'indefArticle': 'a',
                'name': 'compass',
                'order': 5,
                'plural': 'compasses',
                'products': {
                    'edges': [
                        {
                            'node': {
                                'name': 'map1'
                            }
                        },
                        {
                            'node': {
                                'name': 'map2'
                            }
                        },
                        {
                            'node': {
                                'name': 'map3'
                            }
                        }
                    ]
                },
                'singular': 'compass',
                'typeId': '1'
            }
        }
    }
}

snapshots['test_schema_success[update] 2'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'produced_by',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'date_produced',
                                            'type_': 'DATE'
                                        },
                                        'type_': {
                                            'name': 'beam'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-spotlight-beam',
                        'indefArticle': 'a',
                        'name': 'beam',
                        'order': 1,
                        'plural': 'beams',
                        'products': {
                            'edges': [
                            ]
                        },
                        'singular': 'beam',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'fields': {
                            'edges': [
                                {
                                    'node': {
                                        'field': {
                                            'name': 'contact',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'compass'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'field_four',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'compass'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'field': {
                                            'name': 'field_five',
                                            'type_': 'UNICODE_TEXT'
                                        },
                                        'type_': {
                                            'name': 'compass'
                                        }
                                    }
                                }
                            ]
                        },
                        'icon': 'mdi-compass',
                        'indefArticle': 'a',
                        'name': 'compass',
                        'order': 5,
                        'plural': 'compasses',
                        'products': {
                            'edges': [
                                {
                                    'node': {
                                        'name': 'map1'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map2'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map3'
                                    }
                                }
                            ]
                        },
                        'singular': 'compass',
                        'typeId': '1'
                    }
                }
            ],
            'totalCount': 2
        }
    }
}



import datetime
import sqlalchemy
import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product


##__________________________________________________________________||
def test_add(app):
    """A simple test of adding an object with a date field"""

    type1 = ProductType(name="type1")

    # date_produced needs to be initialized with a datetime.date
    date_produced = datetime.date(2019, 2, 23)
    product1 = Product(
        name="product1", date_produced=date_produced, type_=type1
    )

    with app.app_context():
        sa.session.add(product1)
        sa.session.commit()

    with app.app_context():
        product1 = Product.query.filter_by(name="product1").first()

        # The type of the field "date_produced" of Product is "datetime.date"
        assert isinstance(product1.date_produced, datetime.date)

        assert datetime.date(2019, 2, 23) == product1.date_produced


##__________________________________________________________________||
def test_add_raise(app):
    """A simple test of adding an object with a wrong type"""

    type1 = ProductType(name="type1")

    # It is not impossible to instnaiate a date field with a wrong
    # type, e.g, str
    product1 = Product(
        name="product1", date_produced="2019-02-13", type_=type1
    )

    with app.app_context():

        # It is also possible to add
        sa.session.add(product1)

        # However, it is not possible to commit
        with pytest.raises(sqlalchemy.exc.StatementError):
            sa.session.commit()


##__________________________________________________________________||



import pytest
import textwrap

from ....funcs import assert_mutation

HEADERS = {"Authorization": "Bearer token123"}  # octocat


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": textwrap.dedent(
                """
                  mutation m {
                    createProductFilePath(input: {
                      path: "nersc:/go/to/my/new_product_v1",
                      note: "- Note 1",
                      productId: 1010
                    }) { productFilePath { path } }
                  }
                """
            ),
        },
        {
            "query": textwrap.dedent(
                """
                  {
                    product(productId: 1010) {
                      name timePosted producedBy note
                      paths { edges { node { path note product { productId } } } }
                    }
                  }
                """,
            )
        },
        id="createProductFilePath",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_success(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):

    ## To find the token
    # from acondbs.models import GitHubUser
    # with app.app_context():
    #     user1 = GitHubUser.query.filter_by(login='octocat').one()
    #     print(user1)
    #     print(user1.tokens[0].token)

    success = True
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": textwrap.dedent(
                """
                  mutation m {
                    createProductFilePath(input: {
                      path: "nersc:/go/to/my/new_product_v1",
                      note: "- Note 1",
                      productId: 1010,
                      noSuchField: "xxx"
                    }) { productFilePath { path } }
                  }
                """,
            )
        },
        {
            "query": textwrap.dedent(
                """
                  {
                    allProductFilePaths {
                      edges {
                        node {
                          productId
                        }
                      }
                    }
                  }
                """,
            )
        },
        id="createProductFilePath-noSuchField",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_error(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):
    success = False
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[error-immutable-fields] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[delete-paths] 1'] = {
    'data': {
        'updateProduct': {
            'ok': True,
            'product': {
                'attributesBoolean': {
                    'edges': [
                    ]
                },
                'attributesDate': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                },
                                'name': 'date_produced',
                                'value': '2020-02-01'
                            }
                        }
                    ]
                },
                'attributesDateTime': {
                    'edges': [
                    ]
                },
                'attributesFloat': {
                    'edges': [
                    ]
                },
                'attributesInteger': {
                    'edges': [
                    ]
                },
                'attributesTime': {
                    'edges': [
                    ]
                },
                'attributesUnicodeText': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'contact',
                                'value': None
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'produced_by',
                                'value': None
                            }
                        }
                    ]
                },
                'contact': None,
                'dateProduced': '2020-02-01',
                'name': 'map1',
                'note': None,
                'paths': {
                    'edges': [
                    ]
                },
                'postedBy': None,
                'postingGitHubUser': None,
                'producedBy': None,
                'productId': '1',
                'relations': {
                    'edges': [
                        {
                            'node': {
                                'other': {
                                    'name': 'beam1',
                                    'productId': '4',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 4,
                                'relationId': '1',
                                'reverse': {
                                    'relationId': '2',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'parent',
                                        'typeId': '1'
                                    }
                                },
                                'reverseRelationId': 2,
                                'typeId': 2,
                                'type_': {
                                    'name': 'child',
                                    'typeId': '2'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'beam2',
                                    'productId': '5',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 5,
                                'relationId': '3',
                                'reverse': {
                                    'relationId': '4',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'parent',
                                        'typeId': '1'
                                    }
                                },
                                'reverseRelationId': 4,
                                'typeId': 2,
                                'type_': {
                                    'name': 'child',
                                    'typeId': '2'
                                }
                            }
                        }
                    ]
                },
                'timePosted': '2021-01-04T14:32:20',
                'timeUpdated': '2021-01-04T14:32:20',
                'typeId': 1,
                'type_': {
                    'name': 'map',
                    'typeId': '1'
                },
                'updatedBy': 'updater',
                'updatingGitHubUser': {
                    'login': 'user1'
                }
            }
        }
    }
}

snapshots['test_schema_success[delete-paths] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[delete-relations] 1'] = {
    'data': {
        'updateProduct': {
            'ok': True,
            'product': {
                'attributesBoolean': {
                    'edges': [
                    ]
                },
                'attributesDate': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                },
                                'name': 'date_produced',
                                'value': '2020-03-04'
                            }
                        }
                    ]
                },
                'attributesDateTime': {
                    'edges': [
                    ]
                },
                'attributesFloat': {
                    'edges': [
                    ]
                },
                'attributesInteger': {
                    'edges': [
                    ]
                },
                'attributesTime': {
                    'edges': [
                    ]
                },
                'attributesUnicodeText': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'contact',
                                'value': None
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'produced_by',
                                'value': None
                            }
                        }
                    ]
                },
                'contact': None,
                'dateProduced': '2020-03-04',
                'name': 'beam2',
                'note': None,
                'paths': {
                    'edges': [
                        {
                            'node': {
                                'note': None,
                                'path': 'site1:/path/to/beam2',
                                'pathId': '8'
                            }
                        }
                    ]
                },
                'postedBy': None,
                'postingGitHubUser': None,
                'producedBy': None,
                'productId': '5',
                'relations': {
                    'edges': [
                    ]
                },
                'timePosted': '2021-01-04T14:32:20',
                'timeUpdated': '2021-01-04T14:32:20',
                'typeId': 2,
                'type_': {
                    'name': 'beam',
                    'typeId': '2'
                },
                'updatedBy': 'updater',
                'updatingGitHubUser': {
                    'login': 'user1'
                }
            }
        }
    }
}

snapshots['test_schema_success[delete-relations] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[update-paths] 1'] = {
    'data': {
        'updateProduct': {
            'ok': True,
            'product': {
                'attributesBoolean': {
                    'edges': [
                    ]
                },
                'attributesDate': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                },
                                'name': 'date_produced',
                                'value': '2020-02-01'
                            }
                        }
                    ]
                },
                'attributesDateTime': {
                    'edges': [
                    ]
                },
                'attributesFloat': {
                    'edges': [
                    ]
                },
                'attributesInteger': {
                    'edges': [
                    ]
                },
                'attributesTime': {
                    'edges': [
                    ]
                },
                'attributesUnicodeText': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'contact',
                                'value': None
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'produced_by',
                                'value': None
                            }
                        }
                    ]
                },
                'contact': None,
                'dateProduced': '2020-02-01',
                'name': 'map1',
                'note': None,
                'paths': {
                    'edges': [
                        {
                            'node': {
                                'note': None,
                                'path': 'site1:/path/to/map1',
                                'pathId': '1'
                            }
                        },
                        {
                            'node': {
                                'note': None,
                                'path': 'site2:/updated/way/map1',
                                'pathId': '9'
                            }
                        },
                        {
                            'node': {
                                'note': None,
                                'path': 'site4:/additional/map1',
                                'pathId': '10'
                            }
                        }
                    ]
                },
                'postedBy': None,
                'postingGitHubUser': None,
                'producedBy': None,
                'productId': '1',
                'relations': {
                    'edges': [
                        {
                            'node': {
                                'other': {
                                    'name': 'beam1',
                                    'productId': '4',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 4,
                                'relationId': '1',
                                'reverse': {
                                    'relationId': '2',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'parent',
                                        'typeId': '1'
                                    }
                                },
                                'reverseRelationId': 2,
                                'typeId': 2,
                                'type_': {
                                    'name': 'child',
                                    'typeId': '2'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'beam2',
                                    'productId': '5',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 5,
                                'relationId': '3',
                                'reverse': {
                                    'relationId': '4',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'parent',
                                        'typeId': '1'
                                    }
                                },
                                'reverseRelationId': 4,
                                'typeId': 2,
                                'type_': {
                                    'name': 'child',
                                    'typeId': '2'
                                }
                            }
                        }
                    ]
                },
                'timePosted': '2021-01-04T14:32:20',
                'timeUpdated': '2021-01-04T14:32:20',
                'typeId': 1,
                'type_': {
                    'name': 'map',
                    'typeId': '1'
                },
                'updatedBy': 'updater',
                'updatingGitHubUser': {
                    'login': 'user1'
                }
            }
        }
    }
}

snapshots['test_schema_success[update-paths] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/updated/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site4:/additional/map1'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[update-relations] 1'] = {
    'data': {
        'updateProduct': {
            'ok': True,
            'product': {
                'attributesBoolean': {
                    'edges': [
                    ]
                },
                'attributesDate': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                },
                                'name': 'date_produced',
                                'value': '2020-03-04'
                            }
                        }
                    ]
                },
                'attributesDateTime': {
                    'edges': [
                    ]
                },
                'attributesFloat': {
                    'edges': [
                    ]
                },
                'attributesInteger': {
                    'edges': [
                    ]
                },
                'attributesTime': {
                    'edges': [
                    ]
                },
                'attributesUnicodeText': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'contact',
                                'value': None
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'produced_by',
                                'value': None
                            }
                        }
                    ]
                },
                'contact': None,
                'dateProduced': '2020-03-04',
                'name': 'beam2',
                'note': None,
                'paths': {
                    'edges': [
                        {
                            'node': {
                                'note': None,
                                'path': 'site1:/path/to/beam2',
                                'pathId': '8'
                            }
                        }
                    ]
                },
                'postedBy': None,
                'postingGitHubUser': None,
                'producedBy': None,
                'productId': '5',
                'relations': {
                    'edges': [
                        {
                            'node': {
                                'other': {
                                    'name': 'beam1',
                                    'productId': '4',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 4,
                                'relationId': '6',
                                'reverse': {
                                    'relationId': '5',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'child',
                                        'typeId': '2'
                                    }
                                },
                                'reverseRelationId': 5,
                                'typeId': 1,
                                'type_': {
                                    'name': 'parent',
                                    'typeId': '1'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'map2',
                                    'productId': '2',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'map',
                                        'typeId': '1'
                                    }
                                },
                                'otherProductId': 2,
                                'relationId': '8',
                                'reverse': {
                                    'relationId': '7',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'child',
                                        'typeId': '2'
                                    }
                                },
                                'reverseRelationId': 7,
                                'typeId': 1,
                                'type_': {
                                    'name': 'parent',
                                    'typeId': '1'
                                }
                            }
                        }
                    ]
                },
                'timePosted': '2021-01-04T14:32:20',
                'timeUpdated': '2021-01-04T14:32:20',
                'typeId': 2,
                'type_': {
                    'name': 'beam',
                    'typeId': '2'
                },
                'updatedBy': 'updater',
                'updatingGitHubUser': {
                    'login': 'user1'
                }
            }
        }
    }
}

snapshots['test_schema_success[update-relations] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map2'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map2'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[update] 1'] = {
    'data': {
        'updateProduct': {
            'ok': True,
            'product': {
                'attributesBoolean': {
                    'edges': [
                    ]
                },
                'attributesDate': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'date_produced',
                                    'type_': 'DATE'
                                },
                                'name': 'date_produced',
                                'value': '2020-02-01'
                            }
                        }
                    ]
                },
                'attributesDateTime': {
                    'edges': [
                    ]
                },
                'attributesFloat': {
                    'edges': [
                    ]
                },
                'attributesInteger': {
                    'edges': [
                    ]
                },
                'attributesTime': {
                    'edges': [
                    ]
                },
                'attributesUnicodeText': {
                    'edges': [
                        {
                            'node': {
                                'field': {
                                    'name': 'contact',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'contact',
                                'value': 'new-contact'
                            }
                        },
                        {
                            'node': {
                                'field': {
                                    'name': 'produced_by',
                                    'type_': 'UNICODE_TEXT'
                                },
                                'name': 'produced_by',
                                'value': None
                            }
                        }
                    ]
                },
                'contact': 'new-contact',
                'dateProduced': '2020-02-01',
                'name': 'map1',
                'note': '- updated note 123',
                'paths': {
                    'edges': [
                        {
                            'node': {
                                'note': None,
                                'path': 'site1:/path/to/map1',
                                'pathId': '1'
                            }
                        },
                        {
                            'node': {
                                'note': None,
                                'path': 'site2:/another/way/map1',
                                'pathId': '2'
                            }
                        }
                    ]
                },
                'postedBy': None,
                'postingGitHubUser': None,
                'producedBy': None,
                'productId': '1',
                'relations': {
                    'edges': [
                        {
                            'node': {
                                'other': {
                                    'name': 'beam1',
                                    'productId': '4',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 4,
                                'relationId': '1',
                                'reverse': {
                                    'relationId': '2',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'parent',
                                        'typeId': '1'
                                    }
                                },
                                'reverseRelationId': 2,
                                'typeId': 2,
                                'type_': {
                                    'name': 'child',
                                    'typeId': '2'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'beam2',
                                    'productId': '5',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 5,
                                'relationId': '3',
                                'reverse': {
                                    'relationId': '4',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'parent',
                                        'typeId': '1'
                                    }
                                },
                                'reverseRelationId': 4,
                                'typeId': 2,
                                'type_': {
                                    'name': 'child',
                                    'typeId': '2'
                                }
                            }
                        }
                    ]
                },
                'timePosted': '2021-01-04T14:32:20',
                'timeUpdated': '2021-01-04T14:32:20',
                'typeId': 1,
                'type_': {
                    'name': 'map',
                    'typeId': '1'
                },
                'updatedBy': 'updater',
                'updatingGitHubUser': {
                    'login': 'user1'
                }
            }
        }
    }
}

snapshots['test_schema_success[update] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}



import pytest

from .funcs import assert_query

QUERY = '''
{
  __schema {
    types {
      kind
      name
      fields {
        name
      }
    }
    queryType {
      fields {
        name
      }
    }
    mutationType {
      fields {
        name
      }
    }
    subscriptionType {
      fields {
        name
      }
    }
  }
}
'''


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer 90b2ee5fed25506df04fd37343bb68d1803dd97f"},
        id="admin",
    ),
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"},
        id="private",
    ),
    pytest.param(
        {"query": QUERY},
        {"Authorization": "Bearer 1a2d18f270df3abacfb85c5413b668f97794b4ce"},
        id="public-wrong-token",
    ),
    pytest.param(
        {"query": QUERY},
        {},
        id="public-no-token",
    ),
]


@pytest.mark.parametrize("data, headers", params)
@pytest.mark.asyncio
async def test_schema(app_users, snapshot, data, headers):
    await assert_query(app_users, snapshot, data, headers)


##__________________________________________________________________||
